import React, {useCallback, useEffect, useRef, useState} from 'react';
import {BryntumGantt, BryntumGanttProps} from '@bryntum/gantt-react';
import './BGToolbar/BGToolbar';
import {GanttConfig, LocaleManager} from "@bryntum/gantt";
import {Prompt, useLocation} from "react-router";
import {useTranslation} from "react-i18next";
import './BGantt.css'
import './columns/KeyResultColumn'
import './columns/ProgressIconColumn'
import './columns/KeyMilestoneColumn'
import './columns/MilestoneExtColumn'
import './columns/TaskCostCategoryColumn'
import './columns/HyperlinkColumn'
import './columns/HyperlinkAddressColumn'
import {defaultColumns} from "src/main/typescript/components/BGantt/columns/GanttColumns";
import {AccObjectInfo, GToolBar} from "src/main/typescript/components/BGantt/toolbar/GToolBar";
import {GBottomToolBar} from "src/main/typescript/components/BGantt/toolbar/GBottomToolBar";
import {FormBaseUrl} from "src/main/typescript/Utils/ErpConstants";
import {useProjectModelConfigWithAuth} from "src/main/typescript/components/BGantt/GanttUtils";
import {useGetAccObjPlansInfo} from "src/main/typescript/Utils/hooks/useGetAccObjtsPlansInfo";
import {Alert, Spin} from "antd";


export const ProjectGantt = (props: any) => {
    const accObjectId = props.match.params.projectId || props.match.params.versionId
    const baseUrl = props.match.params.projectId ? FormBaseUrl.project : FormBaseUrl.productVersion
    const search = useLocation().search;
    const planId = new URLSearchParams(search).get('planId');
    const {loading: lAccInfo, error: eAccInfo, data: dAccInfo} = useGetAccObjPlansInfo(accObjectId, planId!!)

    const drawContent = () => {
        if (lAccInfo) {
            return <Spin size="large"/>
        }
        if (eAccInfo) {
            return <Alert message={eAccInfo.message} style={{margin: "10px"}}/>
        }
        return <BGantt accObjInfo={dAccInfo}
                       baseUrl={baseUrl}/>
    }
    return (drawContent()
    )
}

export const ProductVersionGantt = (props: any) => {
    return (
        <ProjectGantt {...props} />
    )
}

interface BGanttProps {
    baseUrl: FormBaseUrl
    accObjInfo: AccObjectInfo
}

const BGantt: React.FC<BGanttProps> = ({baseUrl, accObjInfo}) => {
    const {t} = useTranslation();
    const [gantt, setGantt] = useState<any>()
    const ganttRef = useCallback(node => {
        if (node !== null) {
            setGantt(node.instance);
        }
    }, []);
    // @ts-ignore
    const [isReadOnlyPlan, setIsReadOnlyPlan] = useState(accObjInfo.currentPlan?.readonly !== undefined ? accObjInfo.currentPlan?.readonly : true)
    const [isHasChanges, setIsHasChanges] = useState<boolean>(false)

    const taskRenderer = ({taskRecord, renderData}: any) => {
        if (taskRecord.keyMilestone) {
            renderData.wrapperCls += " b-key-milestone"
        }
    }
    const projectConfiguration = useProjectModelConfigWithAuth(accObjInfo.id!!, accObjInfo.currentPlan?.id!!)
    const [ganttConfig, setGanttConfig] = useState({
        rowHeight: 20,
        barMargin: 2,
        dependencyIdField: 'wbsCode',
        subGridConfigs: {
            locked: {
                flex: 3
            },
            normal: {
                flex: 4
            }
        },
        height: "90%",
        loadMaskError: {
            icon: 'b-icon b-icon-warning',
            autoClose: 15000,
            showDelay: 0
        },

        project: projectConfiguration,
        taskRenderer: taskRenderer,
        columns: defaultColumns,
        viewPreset: 'weekAndDayLetter',
    } as Partial<BryntumGanttProps>)

    useEffect(() => {
        window.addEventListener("beforeunload", alertUser)
        return () => {
            window.removeEventListener("beforeunload", alertUser);
            if (gantt) {
                gantt.destroy()
            }
        };
    }, [])

    useEffect(() => {
        if (gantt) {

            if (!accObjInfo.currentPlan?.id) {
                gantt.applyMaskError("Plan not selected")
            }
            // @ts-ignore
            window.gantt = gantt
            // @ts-ignore
            window.LocaleManager = LocaleManager
            gantt.projectId = accObjInfo.id!!
            gantt.selectedPlan = accObjInfo.currentPlan?.id!!
            gantt.project.currentWorkPlan = accObjInfo.workPlan?.id
            // Workaround смотрите RNDTEAM-524, проверить тикет на  gantt после обновления
            gantt.onCrudManagerRequestFail = ({requestType, response, responseText}: any) => {
                gantt.onCrudManagerRequestFinalize(false, requestType, responseText)
            }

            gantt.onBeforeTaskEditShow = (event: any): any => {
                event.taskEdit.editor.widgetMap.taskCostCategory.disabled = !!event.taskRecord.parent.taskCostCategory;
            }
            gantt.onCrudManagerRequestFinalize = (successful: any, requestType: any, responseText: any) => {
                const response = gantt.project.decode(responseText)
                const me = gantt;
                if (successful) {
                    me.masked = null;
                    me.toggleEmptyText?.();
                } else {
                    console.error("Error gantt request failure", response)
                    // Do not remove. Assertion strings for Localization sanity check.
                    // 'L{GridBase.loadFailedMessage}'
                    // 'L{GridBase.syncFailedMessage}'
                    var blob = new Blob([JSON.stringify(response || {})], {type: "text/plain"});
                    var href = window.URL.createObjectURL(blob)
                    var fileName = "error.txt"
                    var onClick = () => {
                        setTimeout(() => {
                            window.URL.revokeObjectURL(href)
                        }, 15000)
                    }
                    me.applyMaskError(
                        `<div class="b-grid-load-failure">
                    <div class="b-grid-load-fail">${me.L(`L{GridBase.${requestType}FailedMessage}`)}</div>
                    ${response && response.message ? `<div class="b-grid-load-fail">${me.L('L{CrudManagerView.serverResponseLabel}')} ${response.message}</div>` : ''}
                    ${t("Report")}: <a download=${fileName} href=${href} onclick=${onClick}>${fileName}</a>
                </div>`);
                }
            }

            gantt.crudManager.on('nochanges', (crud: any) => {
                setIsHasChanges(false)
            })
            gantt.crudManager.on('hasChanges', (crud: any) => {
                setIsHasChanges(true)
            })

            gantt.project.load()
        }
    }, [gantt]);

    useEffect(() => {
        if (gantt) {
            gantt.refresh()
        }
    }, [isReadOnlyPlan])

    const alertUser = (e: any) => {
        if (gantt && gantt.project.crudStoreHasChanges()) {
            e.preventDefault();
            e.returnValue = "";
        }
    };


    return (
        <React.Fragment key={"GanttFragment"}>
            <Prompt
                key={"UnsavedDataWarning"}
                when={isHasChanges}
                message={t('UnsavedDataWarning')}
            />
            <GToolBar isReadOnly={isReadOnlyPlan}
                      key={"GToolBar"}
                      accObject={accObjInfo}
                      baseUrl={baseUrl}
                      ganttRef={gantt}/>
            <BryntumGantt key={"Gantt"} ref={ganttRef}
                          {...ganttConfig}
                          panFeature={{disabled: true}}
                          labelsFeature={{
                              //@ts-ignore
                              disabled: false,
                              left: {
                                  field: 'name',
                                  editor: {
                                      type: 'textfield'
                                  }
                              }
                          }}
                          readOnly={isReadOnlyPlan}
                          dependenciesFeature={{disabled: false}}
                          criticalPathsFeature={{disabled: true}}
                          projectLinesFeature={{disabled: false}}
                          taskEditFeature={{
                              disabled: isReadOnlyPlan,
                              items: {
                                  generalTab: {
                                      items: {
                                          duration: {
                                              flex: '1 0 50%',
                                          },
                                          taskCostCategory: {
                                              // @ts-ignore
                                              store: ganttConfig.project?.taskCostCategoryStore,
                                              type: 'taskcostcategoryfield',
                                              name: "taskCostCategory",
                                              weight: 800,
                                              flex: '1 0 50%',
                                              label: 'L{TaskCostCategory}',

                                          }
                                      }
                                  }
                              }
                          }}
                          rowReorderFeature={
                              {
                                  disabled: isReadOnlyPlan,
                                  listeners: {
                                      gridRowBeforeDropFinalize: (event: any): any => {
                                          if (event.context.parent == event.context.parent.root) {
                                              return false
                                          }
                                      }
                                  }

                              }
                          }
                          cellEditFeature={{disabled: isReadOnlyPlan, addNewAtEnd: false, autoEdit: true}}
                          taskResizeFeature={{disabled: isReadOnlyPlan}}
                          taskDragFeature={{disabled: isReadOnlyPlan}}
                          taskDragCreateFeature={{disabled: isReadOnlyPlan}}
                          taskMenuFeature={{
                              processItems: ({items, taskRecord}) => {
                                  if (taskRecord.parent == taskRecord.stores[0].getRootTask()) {
                                      items.outdent.disabled = true
                                  }
                              },
                              deleteTask: {disabled: isReadOnlyPlan},
                              add: {disabled: isReadOnlyPlan},
                              convertToMilestone: {disabled: isReadOnlyPlan},
                              indent: {
                                  disabled: isReadOnlyPlan
                              },
                              outdent: {
                                  disabled: isReadOnlyPlan

                              },
                          }}
                          filterFeature={true}
            />
            <GBottomToolBar ganttRef={gantt} baseUrl={baseUrl} accObject={accObjInfo}
                            isReadOnly={isReadOnlyPlan}/>
        </React.Fragment>

    );
}
