import {TaskExtModel} from "src/main/typescript/components/BGantt/TaskExtModel";
import {AssignmentModel, DependencyModel, ProjectModel} from "@bryntum/gantt";


interface PlanSynchronizerConfiguration {
    /**
     * Synchronized plan. The plan to which they will be added (Entities from the source will be updated)
     */
    synchronizedProject: ProjectModel,
    /**
     * Data source for synchronization
     */
    sourceProject: ProjectModel
}

export default abstract class AbstractPlanSynchronizer {
    tmpIdFieldName = "$tmpId"
    synchronizedProject: ProjectModel
    sourceProject: ProjectModel
    selectedTasks: Map<string, TaskExtModel>
    selectedAssignments: Array<AssignmentModel>
    selectedDependencies: Array<DependencyModel>

    /**
     * @param props
     * @protected
     */
    constructor(props: PlanSynchronizerConfiguration) {
        this.sourceProject = props.sourceProject
        this.synchronizedProject = props.synchronizedProject
        this.selectedTasks = new Map()
        this.selectedAssignments = []
        this.selectedDependencies = []
        this.init()

    }

    protected init() {
        this.selectTasks().forEach((task) => {
            this.selectedTasks.set(task.id.toString(), task)
            this.selectedAssignments.push(...this.selectAssignments(task))
            this.selectedDependencies.push(...this.selectDependency(task))
        })
        console.debug(this.selectedTasks)
    }

    /**
     * Method defining the list of synchronized tasks
     * @protected
     */
    protected abstract selectTasks(): Array<TaskExtModel>;

    /**
     * A method that determines the rules for deleting entities during synchronization.
     * This method is called after the entities are copied.
     * @protected
     */
    protected abstract clearOldestEntity(): any;

    protected selectAssignments(task: TaskExtModel): Array<AssignmentModel> {
        return task.assignments
    }

    protected selectDependency(task: TaskExtModel): Array<DependencyModel> {
        return task.allDependencies.filter((dependency) => {
            return dependency.toEvent == task.id || dependency.to == task.id
        })
    }

    findNearestParent(selectedTask: TaskExtModel): TaskExtModel {
        let curParent = selectedTask.parent
        // @ts-ignore
        while (curParent && !this.synchronizedProject.taskStore.getByGuid(curParent.getData("guid"))) {
            curParent = curParent.parent
        }
        let result
        if (curParent) {
            // @ts-ignore
            result = this.synchronizedProject.taskStore.getByGuid(curParent.getData("guid"))
        } else {
            // @ts-ignore
            result = this.synchronizedProject.taskStore.getRootTask()
        }
        return result
    }


    prepareSelectedTask(selectedTask: TaskExtModel): TaskExtModel {
        // @ts-ignore
        const copyTask = {...selectedTask.data}
        delete copyTask.id
        if (selectedTask.parent) {
            let curParent = selectedTask.parent
            // @ts-ignore
            let oldParent = this.synchronizedProject.taskStore.getByGuid(curParent.getData("guid"))
            if (oldParent) {
                copyTask.parentId = oldParent.id
            } else if (this.selectedTasks.has(curParent.id.toString())) {
                copyTask.parentId = this.selectedTasks.get(curParent.id.toString())!![this.tmpIdFieldName]
            } else {
                const nearestParent = this.findNearestParent(selectedTask);
                copyTask.parentId = nearestParent?.id
            }
        }
        // @ts-ignore
        const oldTaskByGuid = this.synchronizedProject.taskStore.getByGuid(selectedTask.getData("guid"));
        if (oldTaskByGuid) {
            copyTask.id = oldTaskByGuid.id
        }
        console.debug("CopyTask", copyTask)
        return copyTask
    }

    /**
     * Finds an assignment in the synchronized plan that matches the assignment in the synchronization source.
     * The search is made:
     * 1. By destination guid
     *
     * 2. By linking the GUID of the task and resource
     * @param selectedAssignment Source assignment
     * @return AssignmentModel | undefined if oldest assignment not found
     * @protected
     */
    protected findOldestAssignment(selectedAssignment: AssignmentModel): AssignmentModel {
        // @ts-ignore
        let oldAssignmentByGuid = this.synchronizedProject.assignmentStore.getByGuid(selectedAssignment.getData("guid"))
        if (!oldAssignmentByGuid) {
            // @ts-ignore
            const oldEventByGuid = this.synchronizedProject.taskStore.getByGuid(selectedAssignment.event.guid)
            if (oldEventByGuid) {
                const oldAssignmentsByResourceId = oldEventByGuid.assignments.filter(
                    (itAssignment: AssignmentModel) => itAssignment.resourceId == selectedAssignment.resourceId
                )
                if (oldAssignmentsByResourceId.length > 0) {
                    oldAssignmentByGuid = oldAssignmentsByResourceId[0]
                }
            }
        }
        return oldAssignmentByGuid
    }

    prepareSelectedAssignment(selectedAssignment: AssignmentModel): AssignmentModel {
        // @ts-ignore
        const copyAssignment = {
            ...selectedAssignment.data,
        }
        delete copyAssignment.id
        delete copyAssignment.event
        delete copyAssignment.eventId
        delete copyAssignment.timeSheetIds
        delete copyAssignment.resource
        delete copyAssignment.drawDependencies
        delete copyAssignment.readOnly
        let oldAssignmentByGuid = this.findOldestAssignment(selectedAssignment)
        if (oldAssignmentByGuid) {
            copyAssignment.id = oldAssignmentByGuid.id
            copyAssignment.eventId = oldAssignmentByGuid.event.id
        } else if (selectedAssignment.event[this.tmpIdFieldName]) {
            copyAssignment.eventId = selectedAssignment.event[this.tmpIdFieldName]
        } else {
            const taskByGuid = this.synchronizedProject.taskStore.getByGuid(selectedAssignment.event?.guid)
            if (taskByGuid != null) {
                copyAssignment.eventId = taskByGuid.id
            }
        }
        return copyAssignment
    }

    prepareSelectedDependency(selectedDependency: DependencyModel): DependencyModel | null {
        const preparedDependency = {
            // @ts-ignore
            ...selectedDependency.data,
        }

        delete preparedDependency.id
        delete preparedDependency.from
        delete preparedDependency.fromEvent
        delete preparedDependency.toEvent
        delete preparedDependency.to

        // @ts-ignore
        const fromOldTask = this.synchronizedProject.taskStore.getByGuid(selectedDependency.fromEvent.getData("guid"))
        if (fromOldTask) {
            preparedDependency.from = fromOldTask.id
        } else if (this.selectedTasks.has(selectedDependency.from.toString())) {
            const fromNewTask = this.selectedTasks.get(selectedDependency.from.toString());
            if (fromNewTask && fromNewTask[this.tmpIdFieldName]) {
                preparedDependency.from = fromNewTask[this.tmpIdFieldName]
            }
        } else {
            return null
        }
        if (selectedDependency.toEvent[this.tmpIdFieldName]) {
            preparedDependency.to = selectedDependency.toEvent[this.tmpIdFieldName]
        } else {
            preparedDependency.to = selectedDependency.to
        }
        return preparedDependency;
    }

    synchronize(): Promise<any> {
        this.synchronizedProject.beginBatch()
        this.selectedTasks.forEach((task) => {
            let preparedTask = this.prepareSelectedTask(task)
            if (!preparedTask.id) {
                let addedTask = this.synchronizedProject.taskStore.add(preparedTask)[0]
                task[this.tmpIdFieldName] = addedTask.getData("id")
            } else {
                const updatedTask = this.synchronizedProject.taskStore.getById(preparedTask.id)
                updatedTask.set(preparedTask)
            }
        })
        this.selectedAssignments.forEach((assignment) => {
            let preparedAssignment = this.prepareSelectedAssignment(assignment)
            if (!preparedAssignment.id) {
                this.synchronizedProject.assignmentStore.add(preparedAssignment)
            } else {
                const updatedAssignment = this.synchronizedProject.assignmentStore.getById(preparedAssignment.id)
                updatedAssignment.set(preparedAssignment)
            }
        })
        this.selectedDependencies.forEach((dependency) => {
            let preparedDependency = this.prepareSelectedDependency(dependency)
            if (preparedDependency) {
                this.synchronizedProject.dependencyStore.add(preparedDependency)
            }
        })
        this.clearOldestEntity()
        this.synchronizedProject.endBatch()
        return this.synchronizedProject.commitAsync()
    }

}