import {DateHelper, DependencyModel, Toast} from "@bryntum/gantt";
import {TaskExtModel} from "src/main/typescript/components/BGantt/TaskExtModel";
import DraftWorkPlanSynchronizer
    from "src/main/typescript/components/BGantt/plansynchronizer/DraftWorkPlanSynchronizer";
import {createProjectModel} from "src/main/typescript/components/BGantt/GanttUtils";
import {
    isEveryTasksHaveNoChildren,
    isSomeTasksHaveChildren,
    isTasksFromTheSameHierarchy
} from "src/main/typescript/components/BGantt/toolbar/helpers/GLinkTasksHelpers";

/**
 * Вывод информации на экран
 * @param ganttRef
 * @param message
 */
export const showInfo = (ganttRef: any, message: string) => {
    Toast.show(ganttRef.L(message));
}
/**
 * Проверка, находится ли план в режиме только чтение.
 * Данная проверка должна вызываться во всякой функции, которая изменяет план
 * @param ganttRef
 */
export const isReadOnlyStatus = (ganttRef: any): boolean => {
    if (!ganttRef.selectedPlan) {
        showInfo(ganttRef, 'Select plan');
        return true
    }
    if (ganttRef.readOnly) {
        showInfo(ganttRef, 'Selected plan has readOnly Status');
        return true
    }
    return false
}
/**
 * Проверка на существования ссылки на движок ганта
 * @param ganttRef
 */
export const checkGanttRef = (ganttRef: any): boolean => {
    if (!ganttRef) {
        console.error("Gantt Ref isn't define")
        return false
    }
    return true
}

/**
 * Добавление новой задачи в зависимости от типа задачи
 * @param ganttRef
 * @param type
 */
export const addNewTask = async (ganttRef: any, type: string) => {
    if (!checkGanttRef(ganttRef) || isReadOnlyStatus(ganttRef)) {
        return
    }
    let newTask: Partial<TaskExtModel>
    switch (type) {
        case "summarytask":
            newTask = {
                name: ganttRef.L("New Summary Task"),
            }
            break
        case "milestone":
            newTask = {
                name: ganttRef.L("New Milestone"),
                duration: 0
            }
            break
        case "task":
        default:
            newTask = {
                name: ganttRef.L("New Task"),
                duration: 1
            }
            break
    }
    newTask.schedulingMode = "FixedDuration"
    newTask.effortDriven = true
    let zeroTask = ganttRef.taskStore.getRootTask()
    const added = zeroTask.appendChild(newTask)
    if (type == "summarytask") {
        added.appendChild({
            name: ganttRef.L("New Task"),
            duration: 1
        })
    }
    await ganttRef.project.propagateAsync();

    // scroll to the added task
    await ganttRef.scrollRowIntoView(added);

    ganttRef.features.cellEdit.startEditing({
        record: added,
        field: 'name'
    });
}

/**
 * Вызов карточки редактирования у выбранной задачи
 * @param ganttRef
 */
export const editSelectedTask = (ganttRef: any) => {
    if (!checkGanttRef(ganttRef) || isReadOnlyStatus(ganttRef)) {
        return
    }
    if (ganttRef.selectedRecord) {
        ganttRef.editTask(ganttRef.selectedRecord);
    } else {
        showInfo(ganttRef, 'First select the task you want to edit')

    }
}

/**
 * Удаление выделенных задач
 * @param ganttRef
 */
export const deleteSelectedTasks = (ganttRef: any) => {
    if (!checkGanttRef(ganttRef) || isReadOnlyStatus(ganttRef)) {
        return
    }
    if (ganttRef.selectedRecords.length !== 0) {
        ganttRef.taskStore.remove(ganttRef.selectedRecords)
    } else {
        showInfo(ganttRef, 'First select the task you want to delete')
    }
}

/**
 * Изменение процента завершенности выделенной группы задач
 * @param ganttRef
 * @param percentDone
 */
export const changePercentDoneOnSelectedTasks = async (ganttRef: any, percentDone: number) => {
    if (!checkGanttRef(ganttRef) || isReadOnlyStatus(ganttRef)) {
        return
    }
    ganttRef.project.beginBatch()
    for (let task of ganttRef.selectedRecords) {
        task.percentDone = percentDone
    }
    ganttRef.project.endBatch()
    await ganttRef.project.propagateAsync()
}

/**
 * Приближение масштаба диаграммы
 * @param ganttRef
 */
export const zoomIn = (ganttRef: any) => {
    if (!checkGanttRef(ganttRef)) {
        return
    }
    ganttRef.zoomIn();
}
/**
 * Уменьшение масштаба диаграммы
 * @param ganttRef
 */
export const zoomOut = (ganttRef: any) => {
    if (!checkGanttRef(ganttRef)) {
        return
    }
    ganttRef.zoomOut()
}
/**
 * Центровка масштаба на проекте
 * @param ganttRef
 */
export const zoomToFit = (ganttRef: any) => {
    if (!checkGanttRef(ganttRef)) {
        return
    }
    ganttRef.zoomToFit({
        leftMargin: 50,
        rightMargin: 50
    });
}

/**
 * Включение панорамирования плана (не сочетается с taskDrag и taskDragCreate из-за небольших багов - подтверждено доками)
 * @param ganttRef
 * @param isButtonActive - статус кнопки-переключателя панорамирования на toolbar
 */
export const dragPlan = (ganttRef: any, isButtonActive: boolean) => {
    if (!checkGanttRef(ganttRef)) {
        return
    }
    ganttRef.features.taskDrag.disabled = !ganttRef.readOnly && isButtonActive
    // ganttRef.features.taskDragCreate.disabled = !ganttRef.readOnly && isButtonActive // если понадобится эта доп. фича
    ganttRef.features.pan.disabled = !isButtonActive

}


/**
 * Перевод выделенных задач  в режим ручного планирования
 * @param ganttRef
 */
export const selectedTasksToManuallyScheduled = async (ganttRef: any) => {
    if (!checkGanttRef(ganttRef) || isReadOnlyStatus(ganttRef)) {
        return
    }
    ganttRef.project.beginBatch()
    for (let task of ganttRef.selectedRecords) {
        task.manuallyScheduled = true
    }
    ganttRef.project.endBatch()
    await ganttRef.project.propagateAsync();
}

/**
 * Перевод выделенных задач  в режим автоматического планирования
 * @param ganttRef
 */
export const selectedTasksToAutoScheduled = async (ganttRef: any) => {
    if (!checkGanttRef(ganttRef) || isReadOnlyStatus(ganttRef)) {
        return
    }
    ganttRef.project.beginBatch()
    for (let task of ganttRef.selectedRecords) {
        task.manuallyScheduled = false
    }
    ganttRef.project.endBatch()
    await ganttRef.project.propagateAsync()
}
/**
 * Перевод выделенных задач  в активный режим
 * @param ganttRef
 */
export const selectedTasksToActive = async (ganttRef: any) => {
    if (!checkGanttRef(ganttRef) || isReadOnlyStatus(ganttRef)) {
        return
    }
    ganttRef.project.beginBatch()
    for (let task of ganttRef.selectedRecords) {
        task.inactive = false
    }
    ganttRef.project.endBatch()
    await ganttRef.project.propagateAsync()
}
/**
 * Перевод выделенных задач  в неактивный режим
 * @param ganttRef
 */
export const selectedTasksToInactive = async (ganttRef: any) => {
    if (!checkGanttRef(ganttRef) || isReadOnlyStatus(ganttRef)) {
        return
    }
    ganttRef.project.beginBatch()
    for (let task of ganttRef.selectedRecords) {
        task.inactive = true
    }
    ganttRef.project.endBatch()
    await ganttRef.project.propagateAsync()
}

/**
 * Сброс фильтров
 * @param ganttRef
 */
export const clearTaskStoreFilters = (ganttRef: any) => {
    if (!checkGanttRef(ganttRef)) {
        return
    }
    ganttRef.taskStore.clearFilters()
}

/**
 * Сброс сортировки
 * @param ganttRef
 */
export const clearTaskStoreSorters = (ganttRef: any) => {
    if (!checkGanttRef(ganttRef)) {
        return
    }
    ganttRef.taskStore.sort({ascending: true, columnOwned: true, field: 'wbsValue'})
    ganttRef.taskStore.clearSorters()
}
/**
 * Отправка запроса на сохранение данных
 * @param ganttRef
 */
export const savePlan = (ganttRef: any): Promise<any> => {
    if (!checkGanttRef(ganttRef) || isReadOnlyStatus(ganttRef)) {
        throw Error("Something went wrong")
    }
    return ganttRef.project.sync()
}
/**
 * Синхронизация текущего плана с рабочим плано
 * @param ganttRef
 */
export const syncCurrentPlanWithWork = (ganttRef: any): Promise<any> => {
    if (isReadOnlyStatus(ganttRef) || ganttRef.project.crudStoreHasChanges()) {
        showInfo(ganttRef, "There are unsaved changes")
        throw Error("Something went wrong")
    }
    if (!ganttRef.project.currentWorkPlan) {
        showInfo(ganttRef, "Work plan not found")
        throw Error("Something went wrong")
    }
    ganttRef.masked = {text: ganttRef.L("Synchronization with work plan")}
    let sourceProjectModel = createProjectModel(ganttRef.projectId, ganttRef.project.currentWorkPlan)

    return sourceProjectModel.load().then(() => {
        let synchronizer = new DraftWorkPlanSynchronizer({
            sourceProject: sourceProjectModel,
            synchronizedProject: ganttRef.project
        })
        return synchronizer.synchronize()
    }).catch((info: any) => {
        console.error(info)
        ganttRef.project.revertChanges()
        showInfo(ganttRef, "Something went wrong")
    }).finally(() => {
        ganttRef.masked = false
    })
}
/**
 * Функция заполняет у выбранных задач PercentDone(Завершенность) выделенных в UI задач
 * относительно текущей даты по графику.
 * Т.е завершенность задачи рассчитывается исходя из формулы: (([Текущая дата]-[Дата начала])/([Дата окончания]-[Дата начала]))*100
 * @param ganttRef Ссылка на инициализированный инстанс Gantt
 */
export const onScheduleAction = (ganttRef: any) => {
    let selectedRecords = ganttRef.selectedRecords
    const calculatePercentDone = (task: TaskExtModel) => {
        let startDate = typeof (task.startDate) == "string" ? DateHelper.parse(task.startDate, DateHelper.defaultFormat) : task.startDate
        let endDate = typeof (task.endDate) == "string" ? DateHelper.parse(task.endDate, DateHelper.defaultFormat) : task.endDate
        let now = new Date()
        now.setHours(0, 0, 0, 0);
        let nowStartDatesDiff = DateHelper.diff(startDate, now, "day")
        let endStartDatesDiff = DateHelper.diff(startDate, endDate, "day")
        let percentDone = (((nowStartDatesDiff > 0 ? nowStartDatesDiff : 0) / (endStartDatesDiff ? endStartDatesDiff : 1)) * 100)
        percentDone = percentDone > 100 ? 100 : percentDone
        return percentDone
    }
    ganttRef.taskStore.beginBatch()
    for (const task of selectedRecords) {
        if (!task.children || !task.children.length) {
            task.percentDone = calculatePercentDone(task)
        }
    }
    ganttRef.taskStore.endBatch()
}

/**
 * Функция удаления связей выделенных задач.
 * 1. Если выделена 1 задача, удаляются любые связи данной задачи. Как предшественник, так и последователи.
 * 2. Если выделено больше 1 задачи, удаляются только связи в которых фигурируют обе выделенные задачи.
 * @param ganttRef
 */
export const unlinkSelectedTasks = (ganttRef: any) => {
    if (!checkGanttRef(ganttRef) || isReadOnlyStatus(ganttRef)) {
        return
    }
    let selectedRecords = ganttRef.selectedRecords

    if (selectedRecords?.length == 1) {
        ganttRef.dependencyStore.remove(selectedRecords[0].dependencies)
        return
    }

    let taskMap = new Map(selectedRecords.map((task: TaskExtModel) => {
        return [task.id, task]
    }))
    let willDelete: any = []
    for (const task of selectedRecords) {
        task.predecessors.forEach((dep: DependencyModel) => {

            // @ts-ignore
            if (taskMap.has(dep.toEvent.id) && taskMap.has(dep.fromEvent?.id)) {
                willDelete.push(dep)
            }

        })
    }
    ganttRef.dependencyStore.remove(willDelete)
}
/**
 * Понижает уровень вложенности выделенных задач.
 * Если уровень задачи не может быть понижен, она будет пропущена.
 * @param ganttRef
 */
export const indentSelectedTasks = (ganttRef: any) => {
    if (!checkGanttRef(ganttRef) || isReadOnlyStatus(ganttRef)) {
        return
    }

    const canIndent = (task: TaskExtModel): boolean => {
        return task.id != ganttRef.taskStore.getRootTask(false).id;
    }

    let selectedRecords = ganttRef.selectedRecords
    let willIndent = selectedRecords.filter((task: TaskExtModel) => canIndent(task))
    ganttRef.taskStore.indent(willIndent)
}

/**
 * Повышает уровень вложенности выделенных задач.
 * Если уровень задачи не может быть повышен, она будет пропущена.
 * @param ganttRef
 */
export const outdentSelectedTasks = (ganttRef: any) => {
    if (!checkGanttRef(ganttRef) || isReadOnlyStatus(ganttRef)) {
        return
    }
    const canOutdent = (task: TaskExtModel): boolean => {
        return task.parentId != ganttRef.taskStore.getRootTask(false).id;

    }
    let selectedRecords = ganttRef.selectedRecords
    let willOutdent = selectedRecords.filter((task: TaskExtModel) => canOutdent(task))
    ganttRef.taskStore.outdent(willOutdent)
}
export const linkSelectedTasks = async (ganttRef: any) => {
    if (!checkGanttRef(ganttRef) || isReadOnlyStatus(ganttRef)) {
        return
    }
    let selectedRecords: any[] = ganttRef.selectedRecords
    console.debug("linkSelectedTasks.SelectedRecords", selectedRecords)
    if (selectedRecords.length < 2) {
        return;
    } else if (selectedRecords.length === 2) {
        let leftTask = selectedRecords[0]
        let rightTask = selectedRecords[1]

        if (!isTasksFromTheSameHierarchy(leftTask, rightTask) && await ganttRef.dependencyStore.isValidDependencyToCreate(leftTask, rightTask, 2)) {
            ganttRef.dependencyStore.add([{from: leftTask.id, to: rightTask.id}])
        } else {
            showInfo(ganttRef, 'One task is a subtask of another. Linking is not possible')
        }
    } else if (isEveryTasksHaveNoChildren(selectedRecords)) {
        let leftIndx = 0
        let rightIndx = 1
        const nextPair = () => {
            leftIndx += 1
            rightIndx = leftIndx + 1
        }

        let willCreate = []
        while (leftIndx < selectedRecords.length - 1) {
            let leftTask = selectedRecords[leftIndx]
            let rightTask = selectedRecords[rightIndx]
            let isValidDependency = await ganttRef.dependencyStore.isValidDependencyToCreate(leftTask, rightTask, 2)
            if (isValidDependency) {
                willCreate.push({from: leftTask.id, to: rightTask.id})
            }
            nextPair()
        }
        ganttRef.dependencyStore.add(willCreate)
    } else if (isSomeTasksHaveChildren(selectedRecords)) {
        let willCreate = []

        for(let i = 0; i < selectedRecords.length - 1; i++) {
            for(let j = i + 1; j < selectedRecords.length; j++) {
                const leftTask = selectedRecords[i]
                const rightTask = selectedRecords[j]
                const isValidChildLevels = [leftTask.childLevel, rightTask.childLevel]
                    .every(level => typeof level === 'number');
                let isValidDependency = await ganttRef.dependencyStore.isValidDependencyToCreate(leftTask, rightTask, 2)

                if (isValidChildLevels && isValidDependency) {
                    if (leftTask.childLevel === rightTask.childLevel) {
                        willCreate.push({from: leftTask.id, to: rightTask.id})
                        break
                    } else if (leftTask.childLevel >= rightTask.childLevel) {
                        break;
                    }
                }
            }
        }

        if (willCreate.length) {
            ganttRef.dependencyStore.add(willCreate)
        }
    }
}