import React, {useEffect, useMemo} from "react";
import "./ViewStateSelector.css"
import {
    idGenModes,
    optionItemFunction,
    ViewSelect
} from "src/main/typescript/components/EntityList/ViewSelect";
import {
    CreateGanttViewStateMutationVariables,
    GanttViewStateListAttributesDocument,
    useCreateGanttViewStateMutation, useDeleteGanttViewStateMutation,
    useGanttViewStateListAttributesQuery,
    useGetGanttViewStateMutation, useUpdateGanttViewStateMutation
} from "src/main/typescript/__generated/graphql-types";
import {GanttViewStateType} from "src/main/typescript/components/BGantt/GanttViewStateModel";
import {RouteComponentProps, withRouter} from "react-router";
import {message} from "antd";
import {useTranslation} from "react-i18next";
import {RcFile} from "antd/es/upload";
import {useFileDownload} from "src/main/typescript/Utils/useFile";
import {useGanttStateUtils} from "src/main/typescript/components/BGantt/toolbar/useGanttStateUtils";
import {useGStateContext} from "src/main/typescript/components/BGantt/toolbar/GStateProvider";
import {saveWithCatch} from "src/main/typescript/Utils/ErpUtils";

interface DisplayStateSelectorProps extends RouteComponentProps {
    ganttRef: object,
    objectId: string
}

const customButtons = {
    defaultViewBtn: {
        key: 'defaultViewBtn',
        disabled: true
    }
}

export const ViewStateSelector: React.FC<DisplayStateSelectorProps> = withRouter(({objectId, ganttRef, ...props}) => {
    const accountingObjectId = objectId
    const {t} = useTranslation()
    const {download} = useFileDownload();
    const {setGanttFullState, collectGanttState} = useGanttStateUtils()
    const {setHasGanttChanges} = useGStateContext()

    const initialGanttViewState = collectGanttState(ganttRef, true)

    const refetchGanttViewStateQuery = {
        refetchQueries: [{
            query: GanttViewStateListAttributesDocument, variables: {
                searchCond: "it.accountingObject.$id == '" + accountingObjectId + "'"
            }
        }]
    }

    const {data: listView} = useGanttViewStateListAttributesQuery({
        variables: {searchCond: "it.accountingObject.$id == '" + accountingObjectId + "'"},
        fetchPolicy: "network-only",
        onError: (error: any) => {
            message.error(t('Error'));
            console.error("GraphqlQl Fetch Error", error)
        }
    })

    const [createGanttViewState] = useCreateGanttViewStateMutation(refetchGanttViewStateQuery)
    const [deleteGanttViewState] = useDeleteGanttViewStateMutation(refetchGanttViewStateQuery)
    const [updateGanttViewState] = useUpdateGanttViewStateMutation(refetchGanttViewStateQuery)
    const [getGanttView, {data: ganttView}] = useGetGanttViewStateMutation()


    /**
     *  Наполнение gantt state после выбора представления
     */
    useEffect(() => {
        if (ganttView?.packet?.getGanttViewState?.state) {
            const selectedView = {
                id: ganttView?.packet?.getGanttViewState?.id,
                ...JSON.parse(ganttView?.packet?.getGanttViewState?.state)
            }

            if (selectedView) {
                console.debug('selectedView:', selectedView)
                setGanttFullState(selectedView, ganttRef)
                setHasGanttChanges(true)
            }

        }
    }, [ganttView])

    /**
     * Наполнение списка представлений. Автоматически обновляется после мутаций за счет refetchQuery
     */
    const ganttViewsItems = useMemo(() => {
        let items: any[] = []
        if (listView) {
            // ganttViewStateStore.removeAll()
            listView.searchGanttViewState.elems.forEach((it: any) => {
                items.push({
                    value: it.name,
                    key: it.id
                })
            })
        }
        return items
    }, [listView])


    /**
     * Callbacks для компонента ViewSelect
     */
    /**
     * Выбор представления из Select и подстановка его key/id в мутацию getGanttView
     */
    const selectGanttView: optionItemFunction = (value, key) => {
        if (key) {
            getGanttView({variables: {viewId: key}})
            // ganttRef.current.ganttInstance.selectedViewId = key
            return
        }
        setGanttFullState(initialGanttViewState as GanttViewStateType, ganttRef)
        setHasGanttChanges(true)
    }


    const addGanttView: optionItemFunction = (value, key) => {
        if (value) {
            const viewStateMutationVariables: CreateGanttViewStateMutationVariables = {
                accountingObjectId,
                name: value,
                state: JSON.stringify(collectGanttState(ganttRef))
            }
            saveWithCatch(t, createGanttViewState({variables: viewStateMutationVariables}));
        }
    }

    const deleteGanttView: optionItemFunction = (value, key) => {
        if (key) {
            saveWithCatch(t, deleteGanttViewState({variables: {ganttViewStateId: key}}));
        }
        setGanttFullState(initialGanttViewState as GanttViewStateType, ganttRef)
        setHasGanttChanges(true)
    }

    const updateGanttView: optionItemFunction = (value, key) => {
        saveWithCatch(t, updateGanttViewState({
            variables: {
                ganttViewStateId: key,
                state: JSON.stringify(collectGanttState(ganttRef))
            }
        }));
    }

    const importViewSettings = (file: RcFile) => {
        const isJSON = file.type === "application/json";
        if (!isJSON) {
            message.error(`${t("WrongFileFormat")}: ${file.name}`);
            return;
        }

        const reader = new FileReader();
        reader.readAsText(file);

        reader.onload = () => {
            const fileDataText: string = reader.result as string;
            const dataJSON = JSON.parse(fileDataText);
            const importItemName = dataJSON.items[0].name
            const dublicateItem = ganttViewsItems.filter(item => item.value === importItemName)
            if (dublicateItem.length > 0) {
                saveWithCatch(t, updateGanttViewState({
                    variables: {
                        ganttViewStateId: dublicateItem[0].key,
                        state: dataJSON.items[0].state
                    }
                }));
                return
            }
            const viewStateMutationVariables: CreateGanttViewStateMutationVariables = {
                accountingObjectId,
                name: importItemName,
                state: dataJSON.items[0].state
            }
            saveWithCatch(t, createGanttViewState({variables: viewStateMutationVariables}));
        };
    }

    const exportViewSettings: optionItemFunction = (value, key) => {
        const itemsToExport = listView?.searchGanttViewState.elems.filter((item: any) => item.id === key)
        if (itemsToExport?.length === 0) {
            message.error(t('Error'))
            return
        }
        download(JSON.stringify({items: [itemsToExport?.[0]]}), value, 'json')
    }

    const ganttViewsActionsHandlers = {
        onSelect: selectGanttView,
        onAdd: addGanttView,
        onDelete: deleteGanttView,
        onSave: updateGanttView,
        onImport: importViewSettings,
        onExport: exportViewSettings
    }

    return (
        <ViewSelect items={ganttViewsItems} actions={ganttViewsActionsHandlers} idGenMode={idGenModes.NONE}
                    customButtons={customButtons}/>
    )
})