import {Column, GanttViewStateType} from "src/main/typescript/components/BGantt/GanttViewStateModel";
import {defaultColumns} from "src/main/typescript/components/BGantt/columns/GanttColumns";
import {ColumnStore} from "@bryntum/gantt";

/** helper for gantt columns state restore
 *
 * @param visibleColumnIDs
 * @param currentColumns
 */
const hideGanttColumns = (visibleColumnIDs: string[], currentColumns: Column[]) => {
    return currentColumns.forEach(record => {
        if (!visibleColumnIDs.includes(record.id)) {
            record.hide()
            return
        }
        record.show()
    })
}
/** helper for gantt columns state restore
 *
 * @param savedColumnIDs
 * @param currentColumns
 * @param masterStore
 */
const moveGanttColumns = (savedColumnIDs: string[], currentColumns: Column[], masterStore: ColumnStore) => {
    return savedColumnIDs.forEach(savedColumn => {
        const foundColumn = masterStore.getById(savedColumn)
        if (foundColumn) {
            masterStore.move(foundColumn, masterStore.last)
        }
    })
}

/** helper for gantt columns state restore
 *
 * @param masterStore
 */
const moveGanttColumnsViaStore = (masterStore: ColumnStore) => (savedColumnIDs: string[], currentColumns: Column[]) =>
    moveGanttColumns(savedColumnIDs, currentColumns, masterStore)

export const useGanttStateUtils = () => {

    /**
     * Change full gantt view state: columns, gantt.state
     */
    function setGanttFullState(ganttViewState: GanttViewStateType, ganttRefObject: any) {
        let gantt = ganttRefObject
        if (gantt) {
            if (ganttViewState.state) {
                setGanttState(ganttViewState.state, gantt)
            }


            if (ganttViewState.features) {
                Object.keys(ganttViewState.features).forEach(feature => {
                    if (gantt.features[feature]) {
                        gantt.features[feature].disabled = ganttViewState.features?.[feature]?.disabled ?? false
                    }
                })
            }

            //TODO возможно, для оптимизации в ganttViewState.filters добавить свойство silent:true
            gantt.taskStore.clearFilters()
            if (ganttViewState.filters) {
                gantt.taskStore.filter(ganttViewState.filters)
            }
            gantt.taskStore.sort({ascending: true, columnOwned: true, field: 'wbsValue'})
            gantt.taskStore.clearSorters()
            if (ganttViewState.sorters) {
                gantt.taskStore.sort(ganttViewState.sorters)
            }

            //TODO раскомментить когда будет решение. Смотри RNDTEAM-441.

            if (ganttViewState.columns) {
                // const startTime = performance.now();

                let currentColumns = gantt.columns.records.filter((it: any) => {
                    return it.data.type !== "timeAxis"
                })

                // @Fast
                // const visibleColumnIDs = ganttViewState.columns.map(column => column.id)
                // gantt.columns.beginBatch()
                // hideGanttColumns(visibleColumnIDs, currentColumns)
                // moveGanttColumnsViaStore(gantt.columns)(visibleColumnIDs, currentColumns)
                // gantt.columns.endBatch()

                //@Not fast
                gantt.columns.remove(currentColumns)
                gantt.columns.add(ganttViewState.columns)

                //TODO смотри RNDTEAM-74.

                // const duration = performance.now() - startTime;
                // console.debug(`Setup gantt state took ${duration}ms`);
            }
        }

    }

    /**
     * Change gantt state. Properties State in gantt object
     * Current state will be merge input state.
     * If input state is empty, null or undefined method won't change current state
     * @param ganttState
     * @param ganttRefObject
     */
    function setGanttState(ganttState: object, ganttRefObject: any) {
        if (ganttState) {
            let ganttInstance = ganttRefObject
            if (ganttInstance) {
                ganttInstance.state = {...ganttInstance.state, ...ganttState}
            }
        }
    }

    /**
     * The function returns an object to be saved in the database containing information about the current state of the gantt display
     * The first call can be made after the event mountComponent when the bryntumgantt object is fully formed, i.e. in
     * didComponentMount and later
     * If we call the function before, the function will do nothing.
     */
    function collectGanttState(ganttRefObj: any, isUseDefaults = false) {
        let gantt = ganttRefObj
        if (!gantt) {
            throw new Error('ERROR: GANTT NULLABLE. You may have called the function too early. Read the documentation for the function');
        }
        let ganttViewState: any = {}


        ganttViewState.filters = isUseDefaults ? null : gantt.state?.store?.filters
        ganttViewState.sorters = isUseDefaults ? null : gantt.state?.store?.sorters

        ganttViewState.state = {
            barMargin: isUseDefaults ? 2 : gantt.state.barMargin,
            rowHeight: isUseDefaults ? 20 : gantt.state.rowHeight,
            // zoomLevel: gantt.state.zoomLevel,
            // zoomLevelOptions: gantt.state.zoomLevelOptions,
        }

        ganttViewState.features = {
            dependencies: {disabled: isUseDefaults ? false : gantt.features.dependencies.disabled},
            criticalPaths: {disabled: isUseDefaults ? true : gantt.features.criticalPaths.disabled},
            labels: {disabled: isUseDefaults ? false : gantt.features.labels.disabled}
        }
        //TODO костыль фильтрации сохранения timeAxis. Колонки, которая отвечает за отображение колбасок ганта
        ganttViewState.columns = isUseDefaults ? defaultColumns : gantt.columns.records.filter((it: any) => { // @Not fast
        // ganttViewState.columns = isUseDefaults ? defaultColumns : gantt.columns.visibleColumns.filter((it: any) => { // @Fast
            return it.data.type !== "timeAxis"
        }).map((it: any) => {
            return {
                id: it.data.id,
                type: it.data.type,
                width: it.data.width,
                field: it.data.field,
                hidden: it.data.hidden,
                format: it.data.format
            }
        })

        return ganttViewState
    }


    return {setGanttFullState, collectGanttState}
};
